﻿#region License
// Copyright (c) 2005 Alexander Kolliopoulos
//
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any damages
// arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented; you must not
//    claim that you wrote the original software. If you use this software
//    in a product, an acknowledgment in the product documentation would
//    be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such, and must not
//    be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
#endregion

using System;
using System.Runtime.InteropServices;

namespace Lydos.Matlab
{
    /// <summary>
    /// Encapsulates a Matlab engine connection.
    /// </summary>
    public class Engine : IDisposable
    {
        #region Fields
        IntPtr enginePtr;
        bool bufferEmpty;
        int bufferSize;
        IntPtr bufferPtr;
        #endregion

        #region Constructors
        /// <summary>
        /// Creates a Matlab engine connection with no output buffer.
        /// </summary>
        public Engine() : this(0) { }

        /// <summary>
        /// Creates a Matlab engine connection with the specified output buffer size.
        /// </summary>
        /// <param name="bufferSize">The number of bytes to reserve for the output buffer.</param>
        public Engine(int bufferSize)
        {
            enginePtr = NativeMethods.engOpen(null);
            bufferEmpty = true;
            bufferPtr = IntPtr.Zero;
            BufferSize = bufferSize;
        }
        #endregion

        #region Destructor
        /// <summary>
        /// Closes the connection to Matlab.
        /// </summary>
        public void Close()
        {
            if (Connected)
            {
                Dispose();
            }
        }

        /// <summary>
        /// Closes the connection to Matlab.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Final disposal of object.
        /// </summary>
        /// <param name="disposing">True if called explicitly, false if finalizing.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (Connected)
            {
                NativeMethods.engClose(enginePtr);
                enginePtr = IntPtr.Zero;

                Marshal.FreeHGlobal(bufferPtr);
                bufferPtr = IntPtr.Zero;

                GC.KeepAlive(this);
            }
        }

        /// <summary>
        /// Finalization.
        /// </summary>
        ~Engine()
        {
            Dispose(false);
        }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the size of the result buffer.
        /// </summary>
        /// <value>The size, in bytes, of the result buffer.</value>
        public int BufferSize
        {
            get
            {
                return bufferSize;
            }
            set
            {
                if (bufferSize != value)
                {
                    if (bufferPtr != IntPtr.Zero)
                    {
                        Marshal.FreeHGlobal(bufferPtr);
                    }

                    bufferSize = value;

                    if (bufferSize > 0)
                    {
                        bufferPtr = Marshal.AllocHGlobal(value);
                        NativeMethods.engOutputBuffer(enginePtr, bufferPtr, bufferSize);
                    }
                    else
                    {
                        NativeMethods.engOutputBuffer(enginePtr, IntPtr.Zero, 0);
                    }
                    GC.KeepAlive(this);
                }
            }
        }

        /// <summary>
        /// Gets the state of the Matlab connection.
        /// </summary>
        /// <value>True if there is a valid connection to a Matlab engine, otherwise false.</value>
        public bool Connected
        {
            get
            {
                return enginePtr != IntPtr.Zero;
            }
        }

        /// <summary>
        /// Gets the string result from the last evaluated string.
        /// </summary>
        /// <value>The string result generated by Matlab from the last evaluated string.</value>
        public string Result
        {
            get
            {
                if (bufferEmpty)
                {
                    return String.Empty;
                }
                else
                {
                    return Marshal.PtrToStringAnsi(bufferPtr);
                }
            }
        }

        /// <summary>
        /// Gets or sets the visibility of the Matlab command window.
        /// </summary>
        /// <value>True to show the window, false to hide it.</value>
        /// <exception cref="MatlabUnspecifiedException">Thrown when Matlab returns an error on attempting to get the visibility of the Matlab window.</exception>
        public bool Visible
        {
            get
            {
                bool visible = false;
                int success = NativeMethods.engGetVisible(enginePtr, ref visible);
                GC.KeepAlive(this);

                if (success != 0)
                {
                    throw new MatlabUnspecifiedException("Unable to get visibility status of window.");
                }

                return visible;
            }
            set
            {
                NativeMethods.engSetVisible(enginePtr, value);
                GC.KeepAlive(this);
            }
        }
        #endregion

        #region Methods
        /// <summary>
        /// Evaluates an expression from a string.
        /// </summary>
        /// <param name="expression">A string containing a Matlab expression.</param>
        /// <exception cref="ArgumentNullException">Thrown when <paramref name="expression"/> is null.</exception>
        /// <exception cref="MatlabNotConnectedException">Thrown when there is no open connection to a Matlab engine instance.</exception>
        public void Evaluate(string expression)
        {
            if (expression == null)
            {
                throw new ArgumentNullException("expression");
            }

            if (NativeMethods.engEvalString(enginePtr, expression) != 0)
            {
                throw new MatlabNotConnectedException();
            }
            GC.KeepAlive(this);

            bufferEmpty = false;
        }

        /// <summary>
        /// Evaluates an expression from a string.
        /// </summary>
        /// <param name="expression">A string containing a Matlab expression.</param>
        /// <returns>The result of evaluation.</returns>
        /// <exception cref="ArgumentNullException">Thrown when <paramref name="expression"/> is null.</exception>
        public string EvaluateToString(string expression)
        {
            if (expression == null)
            {
                throw new ArgumentNullException("expression");
            }

            if (BufferSize == 0)
            {
                return null;
            }

            Evaluate(expression);

            return Result;
        }

        /// <summary>
        /// Gets a matrix variable by name.
        /// </summary>
        /// <param name="name">The name of the matrix in Matlab.</param>
        /// <returns>The matrix object.</returns>
        /// <exception cref="ArgumentNullException">Thrown when <paramref name="name"/> is null.</exception>
        /// <exception cref="MatlabGetVariableFailedException">Thrown when the named matrix does not exist.</exception>
        public Matrix GetMatrix(string name)
        {
            if (name == null)
            {
                throw new ArgumentNullException("name");
            }

            IntPtr arrayPtr = NativeMethods.engGetVariable(enginePtr, name);
            GC.KeepAlive(this);

            if (arrayPtr == IntPtr.Zero)
            {
                throw new MatlabGetVariableFailedException();
            }

            return Matrix.FromMxPtr(arrayPtr);
        }

        /// <summary>
        /// Sets a matrix variable by name.
        /// </summary>
        /// <param name="name">The name of the matrix in Matlab.</param>
        /// <param name="matrix">The matrix to store.</param>
        /// <exception cref="ArgumentNullException">Thrown when either <paramref name="name"/> or <paramref name="matrix"/> is null.</exception>
        /// <exception cref="ObjectDisposedException">Thrown when <paramref name="matrix"/> has already been disposed.</exception>
        /// <exception cref="MatlabPutVariableFailedException">Thrown when there is an error returned by Matlab.</exception>
        public void SetMatrix(string name, Matrix matrix)
        {
            if (name == null)
            {
                throw new ArgumentNullException("name");
            }
            if (matrix == null)
            {
                throw new ArgumentNullException("matrix");
            }
            if (matrix.MxPtr == IntPtr.Zero)
            {
                throw new ObjectDisposedException("matrix", "Cannot operate on disposed matrix.");
            }

            int success = NativeMethods.engPutVariable(enginePtr, name, matrix.MxPtr);
            GC.KeepAlive(matrix);
            GC.KeepAlive(this);

            if (success != 0)
            {
                throw new MatlabPutVariableFailedException();
            }
        }

        /// <summary>
        /// Creates and returns a string indicating the state of the Matlab connection.
        /// </summary>
        /// <returns>The state of the connection, or the result buffer contents, if there is any.</returns>
        public override string ToString()
        {
            if (enginePtr == IntPtr.Zero)
            {
                return "Not connected";
            }
            else if (BufferSize == 0)
            {
                return "Zero size result buffer";
            }
            else
            {
                return Result;
            }
        }
        #endregion
    }
}
