#region License
// Copyright (c) 2005 Alexander Kolliopoulos
//
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any damages
// arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented; you must not
//    claim that you wrote the original software. If you use this software
//    in a product, an acknowledgment in the product documentation would
//    be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such, and must not
//    be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
#endregion

using System;

namespace Lydos.Matlab
{
    /// <summary>
    /// Provides access to a Matlab MAT file.
    /// </summary>
    public class MatFile : IDisposable
    {
        #region Fields
        private IntPtr matFilePtr;
        #endregion

        #region Enums
        /// <summary>
        /// Read/write modes for accessing MAT files.
        /// </summary>
        public enum OpenMode 
        {
            /// <summary>
            /// Read only.
            /// </summary>
            Read,

            /// <summary>
            /// Read and write, but not create.
            /// </summary>
            Update,

            /// <summary>
            /// Overwrite.
            /// </summary>
            Write,

            /// <summary>
            /// Write MATLAB 4 (and earlier) compatible file.
            /// </summary>
            WriteLevel4,

            /// <summary>
            /// Write character data with system character set.
            /// </summary>
            WriteSysCharSet,

            /// <summary>
            /// Write compressed data.
            /// </summary>
            WriteCompressed
        }
        #endregion

        #region Constructors
        /// <summary>
        /// Opens a MAT file as read only.
        /// </summary>
        /// <param name="fileName">The name of the file to open.</param>
        public MatFile(string fileName) : this(fileName, OpenMode.Read) { }

        /// <summary>
        /// Opens a MAT file in the specified mode.
        /// </summary>
        /// <param name="fileName">The name of the file to open.</param>
        /// <param name="mode">The mode to open the file in.</param>
        /// <exception cref="MatlabUnspecifiedException">Thrown when there is an error opening the file.</exception>
        public MatFile(string fileName, OpenMode mode) 
        {
            string strMode = String.Empty;

            switch (mode)
            {
                case OpenMode.Read:
                    strMode = "r";
                    break;

                case OpenMode.Update:
                    strMode = "u";
                    break;

                case OpenMode.Write:
                    strMode = "w";
                    break;

                case OpenMode.WriteLevel4: 
                    strMode = "w4";
                    break;

                case OpenMode.WriteSysCharSet:
                    strMode = "wL";
                    break;

                case OpenMode.WriteCompressed:
                    strMode = "wz";
                    break;
            }

            matFilePtr = NativeMethods.matOpen(fileName, strMode);

            if (matFilePtr == IntPtr.Zero)
            {
                throw new MatlabUnspecifiedException("MAT file open failed");
            }
            GC.KeepAlive(this);
        }
        #endregion

        #region Destructor
        /// <summary>
        /// Closes the MAT file.
        /// </summary>
        public void Close()
        {
            if (IsOpen)
            {
                Dispose();
            }
        }

        /// <summary>
        /// Closes the MAT file.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Final disposal of object.
        /// </summary>
        /// <param name="disposing">True if called explicitly, false if finalizing.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (matFilePtr != IntPtr.Zero)
            {
                NativeMethods.matClose(matFilePtr);
                GC.KeepAlive(this);
                matFilePtr = IntPtr.Zero;
            }
        }

        /// <summary>
        /// Finalization.
        /// </summary>
        ~MatFile()
        {
            Dispose(false);
        }
        #endregion

        #region Properties
        /// <summary>
        /// Gets whether or not a MAT file is open.
        /// </summary>
        /// <value>True if a MAT file is open, otherwise false.</value>
        public bool IsOpen
        {
            get
            {
                return matFilePtr != IntPtr.Zero;
            }
        }
        #endregion

        #region Methods
        /// <summary>
        /// Deletes a variable.
        /// </summary>
        /// <param name="name">The name of the variable to delete.</param>
        /// <exception cref="ArgumentNullException">Thrown when <paramref name="name"/> is null.</exception>
        /// <exception cref="ObjectDisposedException">Thrown when this has already been disposed.</exception>
        /// <exception cref="MatlabDeleteVariableFailedException">Thrown when the specified variable name does not exist or cannot be deleted.</exception>
        public void DeleteVariable(string name)
        {
            if (name == null)
            {
                throw new ArgumentNullException("name");
            }
            if (matFilePtr == IntPtr.Zero)
            {
                throw new ObjectDisposedException(null, "Cannot access disposed MAT file.");
            }

            if (NativeMethods.matDeleteVariable(matFilePtr, name) != 0)
            {
                throw new MatlabDeleteVariableFailedException();
            }

            GC.KeepAlive(this);
        }

        /// <summary>
        /// Gets a variable.
        /// </summary>
        /// <param name="name">The name of the variable to get.</param>
        /// <returns>The specified variable.</returns>
        /// <exception cref="ArgumentNullException">Thrown when <paramref name="name"/> is null.</exception>
        /// <exception cref="ObjectDisposedException">Thrown when this has already been disposed.</exception>
        /// <exception cref="MatlabGetVariableFailedException">Thrown when the specified variable does not exist.</exception>
        public Matrix GetVariable(string name)
        {
            if (name == null)
            {
                throw new ArgumentNullException("name");
            }
            if (matFilePtr == IntPtr.Zero) 
            {
                throw new ObjectDisposedException(null, "Cannot access disposed MAT file.");
            }

            IntPtr mxArray = NativeMethods.matGetVariable(matFilePtr, name);
            GC.KeepAlive(this);

            if (mxArray == IntPtr.Zero)
            {
                throw new MatlabGetVariableFailedException();
            }

            return Matrix.FromMxPtr(mxArray);
        }

        /// <summary>
        /// Sets a variable.
        /// </summary>
        /// <param name="name">The name of the variable to set.</param>
        /// <param name="var">The matrix to store in the variable.</param>
        /// <exception cref="ArgumentNullException">Thrown when <paramref name="name"/> or <paramref name="var"/> is null.</exception>
        /// <exception cref="ObjectDisposedException">Thrown when <paramref name="var"/> or this has already been disposed.</exception>
        /// <exception cref="MatlabPutVariableFailedException">Thrown when there is an error setting the variable.</exception>
        public void PutVariable(string name, Matrix var)
        {
            if (name == null)
            {
                throw new ArgumentNullException("name");
            }
            if (var == null)
            {
                throw new ArgumentNullException("var");
            }
            if (matFilePtr == IntPtr.Zero)
            {
                throw new ObjectDisposedException(null, "Cannot access disposed MAT file.");
            }
            if (var.MxPtr == IntPtr.Zero)
            {
                throw new ObjectDisposedException("var", "Cannot operate on disposed matrix.");
            }

            int success = NativeMethods.matPutVariable(matFilePtr, name, var.MxPtr);
            GC.KeepAlive(var);
            GC.KeepAlive(this);

            if (success != 0)
            {
                throw new MatlabPutVariableFailedException();
            }
        }

        /// <summary>
        /// Sets a variable as global.
        /// </summary>
        /// <param name="name">The name of the variable to set.</param>
        /// <param name="var">The matrix to store in the variable.</param>
        /// <exception cref="ArgumentNullException">Thrown when <paramref name="name"/> or <paramref name="var"/> is null.</exception>
        /// <exception cref="ObjectDisposedException">Thrown when <paramref name="var"/> or this has already been disposed.</exception>
        /// <exception cref="MatlabPutVariableFailedException">Thrown when there is an error setting the variable.</exception>
        public void PutVariableAsGlobal(string name, Matrix var) 
        {
            if (name == null)
            {
                throw new ArgumentNullException("name");
            }
            if (var == null)
            {
                throw new ArgumentNullException("var");
            }
            if (matFilePtr == IntPtr.Zero)
            {
                throw new ObjectDisposedException(null, "Cannot access disposed MAT file.");
            }
            if (var.MxPtr == IntPtr.Zero)
            {
                throw new ObjectDisposedException("var", "Cannot operate on disposed matrix.");
            }

            int success = NativeMethods.matPutVariableAsGlobal(matFilePtr, name, var.MxPtr);
            GC.KeepAlive(var);
            GC.KeepAlive(this);

            if (success != 0)
            {
                throw new MatlabPutVariableFailedException();
            }
        }

        /// <summary>
        /// Gets a list of variable names available in the MAT file.
        /// </summary>
        /// <returns>An array of strings corresponding to variable names.</returns>
        /// <exception cref="ObjectDisposedException">Thrown when this has already been disposed.</exception>
        /// <exception cref="MatlabUnspecifiedException">Thrown when there is an error getting the variable list, such as when the MAT file is not opened in a readable mode.</exception>
        public string[] VariableList()
        {
            if (matFilePtr == IntPtr.Zero) 
            {
                throw new ObjectDisposedException(null, "Cannot access disposed MAT file.");
            }

            int count;
            IntPtr mxPtr = NativeMethods.matGetDir(matFilePtr, out count);

            if (mxPtr == IntPtr.Zero)
            {
                throw new MatlabUnspecifiedException();
            }

            string[] list = new string[count];

            unsafe
            {
                char** p = (char**)mxPtr.ToPointer();
                for (int i = 0; i < count; ++i)
                {
                    list[i] = System.Runtime.InteropServices.Marshal.PtrToStringAnsi((IntPtr)p[i]);
                }
            }

            NativeMethods.mxFree(mxPtr);
            GC.KeepAlive(this);

            return list;
        }
        #endregion
    }
}
